#ifdef PETSC_RCS_HEADER
static char vcid[] = "$Id: heat.c,v 1.5 2000/02/01 17:02:51 knepley Exp $";
#endif

static char help[] = "This is the Heat equation with Dirchlet boundary conditions.\n\n";

int HEAT_COOKIE;
int HEAT_ComputeField;

#include "heat.h"

/*
  Here we are solving the Heat equation:

    du/dt - \Delta u = f(x)

  Thus we now let

    SolutionFunction() --> u_S (which we prescribe)
    RhsFunction()      --> f
*/
#undef  __FUNCT__
#define __FUNCT__ "main"
int main(int argc, char **args) {
  HeatContext ctx; /* Holds problem specific information */
  int         ierr;

  PetscFunctionBegin;
  ierr = PetscInitialize(&argc, &args, 0, help);                                    CHKERRABORT(PETSC_COMM_WORLD, ierr);

  ierr = HeatInitializePackage(PETSC_NULL);                                         CHKERRABORT(PETSC_COMM_WORLD, ierr);
  ierr = HeatContextCreate(PETSC_COMM_WORLD, &ctx);                                 CHKERRABORT(PETSC_COMM_WORLD, ierr);
  ierr = HeatComputeField(ctx);                                                     CHKERRABORT(PETSC_COMM_WORLD, ierr);
  ierr = HeatContextDestroy(ctx);                                                   CHKERRABORT(PETSC_COMM_WORLD, ierr);

  CHKMEMQ;
  PetscFinalize();
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "HeatInitializePackage"
/*@C
  HeatInitializePackage - This function initializes everything in the Heat package.

  Input Parameter:
. path - The dynamic library path, or PETSC_NULL

  Level: developer

.keywords: Heat, initialize, package
.seealso: PetscInitialize()
@*/
int HeatInitializePackage(char *path) {
  static PetscTruth initialized = PETSC_FALSE;
  char              logList[256];
  char             *className;
  PetscTruth        opt;
  int               ierr;

  PetscFunctionBegin;
  if (initialized == PETSC_TRUE) PetscFunctionReturn(0);
  initialized = PETSC_TRUE;
  /* Register Classes */
  ierr = PetscLogClassRegister(&HEAT_COOKIE, "Heat");                                                     CHKERRQ(ierr);
  /* Register Constructors and Serializers */
  /* Register Events */
  ierr = PetscLogEventRegister(&HEAT_ComputeField, "HeatComputeField", HEAT_COOKIE);                      CHKERRQ(ierr);
  /* Process info exclusions */
  ierr = PetscOptionsGetString(PETSC_NULL, "-log_info_exclude", logList, 256, &opt);                      CHKERRQ(ierr);
  if (opt == PETSC_TRUE) {
    ierr = PetscStrstr(logList, "HeatContext", &className);                                               CHKERRQ(ierr);
    if (className) {
      ierr = PetscLogInfoDeactivateClass(HEAT_COOKIE);                                                    CHKERRQ(ierr);
    }
  }
  /* Process summary exclusions */
  ierr = PetscOptionsGetString(PETSC_NULL, "-log_summary_exclude", logList, 256, &opt);                   CHKERRQ(ierr);
  if (opt == PETSC_TRUE) {
    ierr = PetscStrstr(logList, "HeatContext", &className);                                               CHKERRQ(ierr);
    if (className) {
      ierr = PetscLogEventDeactivateClass(HEAT_COOKIE);                                                   CHKERRQ(ierr);
    }
  }
  PetscFunctionReturn(0);
}

/*-------------------------------------------- HeatContext Creation ------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "HeatContextCreate"
/*@
  HeatContextCreate - This function initializes the Heat context.

  Collective on MPI_Comm

  Input Parameter:
. comm - The communicator

  Output Parameter:
. sCtx - The HeatContext

  Level: beginner

.keywords: Heat, context, create
.seealso: HeatContextDestroy(), HeatContextPrint(), HeatContextSetup()
@*/
int HeatContextCreate(MPI_Comm comm, HeatContext *sCtx) {
  HeatContext ctx;

  PetscFunctionBegin;
  /* Setup context */
  PetscHeaderCreate(ctx, _HeatContext, int, HEAT_COOKIE, -1, "HeatContext", comm, 0, 0);
  PetscLogObjectCreate(ctx);
  PetscLogObjectMemory(ctx, sizeof(struct _HeatContext));

  /* Initialize subobjects */
  ctx->grid   = PETSC_NULL;
  ctx->ts     = PETSC_NULL;
  ctx->u      = PETSC_NULL;
  ctx->uExact = PETSC_NULL;
  /* Setup domain */
  ctx->geometryCtx.size[0]  = 2.0;
  ctx->geometryCtx.size[1]  = 2.0;
  ctx->geometryCtx.start[0] = 0.0;
  ctx->geometryCtx.start[1] = 0.0;
  /* Setup refinement */
  ctx->geometryCtx.maxArea  = 0.5;
  ctx->geometryCtx.areaFunc = PointFunctionConstant;
  ctx->geometryCtx.areaCtx  = PETSC_NULL;
  /* Initialize problem loop */
  ctx->dim        = 2;
  ctx->linear     = PETSC_FALSE;
  ctx->refineStep = 0;
  ctx->numLoops   = 0;
  /* Initialize physics */
  ctx->decayExponent = 1.0;

  *sCtx = ctx;
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatContextDestroy"
/*@
  HeatContextDestroy - This function destroys the Heat context.

  Collective on HeatContext

  Input Parameter:
. ctx - The HeatContext

  Level: beginner

.keywords: Heat, context, destroy
.seealso: HeatContextCreate(), HeatContextSetup()
@*/
int HeatContextDestroy(HeatContext ctx) {
  Grid grid = ctx->grid;
  int  ierr;

  PetscFunctionBegin;
  if (--ctx->refct > 0) SETERRQ(PETSC_ERR_PLIB, "Heat context should not be referenced more than once");
  ierr = GridFinalizeBC(grid);                                                                            CHKERRQ(ierr);
  ierr = GridDestroy(grid);                                                                               CHKERRQ(ierr);
  PetscLogObjectDestroy(ctx);
  PetscHeaderDestroy(ctx);
  PetscFunctionReturn(0);
}

/*--------------------------------------------- HeatContext Setup -------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "HeatContextSetup"
/*@
  HeatContextSetup - This function configures the context from user options.

  Collective on HeatContext

  Input Parameter:
. ctx - The HeatContext

  Options Database Keys:
+ dim <num>           - The problem dimension
. linear              - Use a linear discretization
. num_systems <num>   - The number of systems to solve
. refine_step <num>   - The step to start refining the mesh
. mesh_max_area <num> - The maximum area of a triangle in the refined mesh
- decay_exponent <k>  - The exponent k in our exact solution

  Level: beginner

.seealso HeatRefineGrid(), HeatDestroyGrid()
*/
int HeatContextSetup(HeatContext ctx) {
  PetscTruth opt;
  int        ierr;

  PetscFunctionBegin;
  /* Determine the problem dimension */
  ierr = PetscOptionsGetInt(PETSC_NULL, "-dim", &ctx->dim, &opt);                                         CHKERRQ(ierr);
  /* Determine the element type */
  ierr = PetscOptionsHasName(PETSC_NULL, "-linear", &ctx->linear);                                        CHKERRQ(ierr);
  /* Determine how many systems to solve */
  ierr = PetscOptionsGetInt(PETSC_NULL, "-num_systems", &ctx->numLoops, &opt);                            CHKERRQ(ierr);
  /* The first iteration at which to refine the mesh */
  ierr = PetscOptionsGetInt(PETSC_NULL, "-refine_step", &ctx->refineStep, &opt);                          CHKERRQ(ierr);
  /* The maximum area of any triangle in the refined mesh */
  ierr = PetscOptionsGetReal("mesh_", "-max_area", &ctx->geometryCtx.maxArea, &opt);                      CHKERRQ(ierr);
  /* The maximum area of any triangle in the refined mesh */
  ierr = PetscOptionsGetScalar(PETSC_NULL, "-decay_exponent", &ctx->decayExponent, &opt);                 CHKERRQ(ierr);

  /* Create main problem */
  ierr = HeatContextCreateGrid(ctx);                                                                      CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/*------------------------------------------------ Grid Creation -----------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "HeatContextCreateMeshBoundary"
/*@
  HeatContextCreateMeshBoundary - This function creates a mesh boundary for the main problem.

  Collective on HeatContext

  Input Parameter:
. ctx - A HeatContext with problem specific information

  Level: beginner

.seealso HeatContextDestroyMeshBoundary(), HeatContextCreateMesh()
@*/
int HeatContextCreateMeshBoundary(HeatContext ctx) {
  MPI_Comm             comm;
  MeshGeometryContext *geomCtx = &ctx->geometryCtx;
  int                  ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject) ctx, &comm);                                                    CHKERRQ(ierr);
  switch(ctx->dim) {
  case 1:
    ierr = MeshBoundary1DCreateSimple(comm, geomCtx, &ctx->boundaryCtx);                                  CHKERRQ(ierr);
    break;
  case 2:
    ierr = MeshBoundary2DCreateSimple(comm, geomCtx, &ctx->boundaryCtx);                                  CHKERRQ(ierr);
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatContextDestroyMeshBoundary"
/*@
  HeatContextDestroyMeshBoundary - This function destroys the mesh boundary for the main problem.

  Collective on HeatContext

  Input Parameter:
. ctx - A HeatContext with problem specific information

  Level: beginner

.seealso HeatContextCreateMeshBoundary(), HeatContextCreateMesh()
@*/
int HeatContextDestroyMeshBoundary(HeatContext ctx) {
  MPI_Comm comm;
  int      ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject) ctx, &comm);                                                    CHKERRQ(ierr);
  switch(ctx->dim) {
  case 1:
    ierr = MeshBoundary1DDestroy(comm, &ctx->boundaryCtx);                                                CHKERRQ(ierr);
    break;
  case 2:
    ierr = MeshBoundary2DDestroy(comm, &ctx->boundaryCtx);                                                CHKERRQ(ierr);
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatContextCreateMesh"
/*@
  HeatContextCreateMesh - This function creates a mesh for the main problem.

  Collective on HeatContext

  Input Parameter:
. ctx - A HeatContext with problem specific information

  Output Parameter:
. m   - The Mesh

  Options Database Keys:
. mesh_refine   - Refines the mesh based on area criteria
. mesh_max_area - The maximum area of an element

  Level: beginner

.seealso HeatRefineGrid(), HeatDestroyGrid()
@*/
int HeatContextCreateMesh(HeatContext ctx, Mesh *m) {
  MeshGeometryContext *geomCtx = &ctx->geometryCtx;
  MPI_Comm             comm;
  Mesh                 mesh;
  Partition            part;
  int                  totalElements, totalNodes, totalEdges;
  char                 name[1024];
  int                  d;
  int                  ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject) ctx, &comm);                                                    CHKERRQ(ierr);
  ierr = HeatContextCreateMeshBoundary(ctx);                                                              CHKERRQ(ierr);
  ierr = MeshCreate(comm, &mesh);                                                                         CHKERRQ(ierr);
  ierr = MeshSetDimension(mesh, ctx->dim);                                                                CHKERRQ(ierr);
  for(d = 0; d < ctx->dim; d++) {
    ierr = MeshSetPeriodicDimension(mesh, d, geomCtx->isPeriodic[d]);                                     CHKERRQ(ierr);
  }
  switch(ctx->dim) {
  case 1:
    if (ctx->linear == PETSC_TRUE) {
      ierr = MeshSetNumCorners(mesh, 2);                                                                  CHKERRQ(ierr);
    } else {
      ierr = MeshSetNumCorners(mesh, 3);                                                                  CHKERRQ(ierr);
    }
    break;
  case 2:
    if (ctx->linear == PETSC_TRUE) {
      ierr = MeshSetNumCorners(mesh, 3);                                                                  CHKERRQ(ierr);
    } else {
      ierr = MeshSetNumCorners(mesh, 6);                                                                  CHKERRQ(ierr);
    }
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }
  ierr = MeshSetBoundary(mesh, &ctx->boundaryCtx);                                                        CHKERRQ(ierr);
  sprintf(name, "mesh.r%.4g", geomCtx->maxArea);
  ierr = PetscObjectSetName((PetscObject) mesh, name);                                                    CHKERRQ(ierr);
  ierr = MeshSetFromOptions(mesh);                                                                        CHKERRQ(ierr);
  ierr = HeatContextDestroyMeshBoundary(ctx);                                                             CHKERRQ(ierr);
  /* Setup refinement */
  if (ctx->geometryCtx.areaCtx != PETSC_NULL) {
    ierr = MeshSetUserContext(mesh, geomCtx->areaCtx);                                                    CHKERRQ(ierr);
  } else {
    ierr = MeshSetUserContext(mesh, &geomCtx->maxArea);                                                   CHKERRQ(ierr);
  }
  /* Report on mesh */
  ierr = MeshGetPartition(mesh, &part);                                                                   CHKERRQ(ierr);
  switch(ctx->dim) {
  case 1:
    ierr = PartitionGetTotalElements(part, &totalElements);                                               CHKERRQ(ierr);
    ierr = PartitionGetTotalNodes(part, &totalNodes);                                                     CHKERRQ(ierr);
    PetscPrintf(comm, "Elements: %d Nodes: %d\n", totalElements, totalNodes);
    break;
  case 2:
    ierr = PartitionGetTotalElements(part, &totalElements);                                               CHKERRQ(ierr);
    ierr = PartitionGetTotalNodes(part, &totalNodes);                                                     CHKERRQ(ierr);
    ierr = PartitionGetTotalEdges(part, &totalEdges);                                                     CHKERRQ(ierr);
    PetscPrintf(comm, "Elements: %d Nodes: %d Edges: %d\n", totalElements, totalNodes, totalEdges);
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }

  *m = mesh;
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatContextCreateGrid"
/*@
  HeatContextCreateGrid - This function creates a grid for the main problem.

  Collective on HeatContext

  Input Parameter:
. ctx     - A HeatContext with problem specific information

  Level: beginner

.seealso HeatRefineGrid(), HeatDestroyGrid()
@*/
int HeatContextCreateGrid(HeatContext ctx) {
  Mesh mesh;
  Grid grid;
  int  ierr;

  PetscFunctionBegin;
  /* Construct the mesh */
  ierr = HeatContextCreateMesh(ctx, &mesh);                                                               CHKERRQ(ierr);
  /* Construct the grid */
  ierr = GridCreate(mesh, &grid);                                                                         CHKERRQ(ierr);
  ierr = MeshDestroy(mesh);                                                                               CHKERRQ(ierr);
  switch(ctx->dim) {
  case 1:
    if (ctx->linear == PETSC_TRUE) {
      ierr = GridAddField(grid, "u", DISCRETIZATION_TRIANGULAR_1D_LINEAR, 1, PETSC_NULL);                 CHKERRQ(ierr);
    } else {
      ierr = GridAddField(grid, "u", DISCRETIZATION_TRIANGULAR_1D_QUADRATIC, 1, PETSC_NULL);              CHKERRQ(ierr);
    }
    break;
  case 2:
    if (ctx->linear == PETSC_TRUE) {
      ierr = GridAddField(grid, "u", DISCRETIZATION_TRIANGULAR_2D_LINEAR, 2, PETSC_NULL);                 CHKERRQ(ierr);
    } else {
      ierr = GridAddField(grid, "u", DISCRETIZATION_TRIANGULAR_2D_QUADRATIC, 2, PETSC_NULL);              CHKERRQ(ierr);
    }
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }
  ierr = GridSetFromOptions(grid);                                                                        CHKERRQ(ierr);

  ctx->grid = grid;
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatRefineGrid"
/*@
  HeatRefineGrid - This function refines the mesh for the main grid.

  Collective on HeatContext

  Input Parameters:
+ ctx - A HeatContext

  Options Database Keys:
. mesh_max_area - The maximum area an element may have

  Level: beginner

.seealso HeatCreateGrid(), HeatDestroyGrid()
@*/
int HeatRefineGrid(HeatContext ctx) {
  Grid                 oldGrid = ctx->grid;
  Grid                 grid;
  Mesh                 mesh;
  Partition            part;
  MeshGeometryContext *geomCtx = &ctx->geometryCtx;
  char                 name[1024];
  int                  totalElements, totalNodes, totalEdges;
  int                  ierr;

  PetscFunctionBegin;
  /* Construct a refined mesh */
  if (geomCtx->areaCtx != PETSC_NULL) {
    ierr = GridRefineMesh(oldGrid, geomCtx->areaFunc, &grid);                                             CHKERRQ(ierr);
    ierr = GridGetMesh(grid, &mesh);                                                                      CHKERRQ(ierr);
    ierr = MeshSetUserContext(mesh, geomCtx->areaCtx);                                                    CHKERRQ(ierr);
  } else {
    geomCtx->maxArea *= 0.5;
    ierr = GridRefineMesh(oldGrid, geomCtx->areaFunc, &grid);                                             CHKERRQ(ierr);
    ierr = GridGetMesh(grid, &mesh);                                                                      CHKERRQ(ierr);
    ierr = MeshSetUserContext(mesh, &geomCtx->maxArea);                                                   CHKERRQ(ierr);
  }
  sprintf(name, "mesh.r%.4g", geomCtx->maxArea);
  ierr = PetscObjectSetName((PetscObject) mesh, name);                                                    CHKERRQ(ierr);
  ierr = MeshSetOptionsPrefix(mesh, "ref_");                                                              CHKERRQ(ierr);
  ierr = GridSetOptionsPrefix(grid, "ref_");                                                              CHKERRQ(ierr);
  ierr = GridSetFromOptions(grid);                                                                        CHKERRQ(ierr);

  ierr = MeshGetPartition(mesh, &part);                                                                   CHKERRQ(ierr);
  ierr = PartitionGetTotalElements(part, &totalElements);                                                 CHKERRQ(ierr);
  ierr = PartitionGetTotalNodes(part, &totalNodes);                                                       CHKERRQ(ierr);
  ierr = PartitionGetTotalEdges(part, &totalEdges);                                                       CHKERRQ(ierr);
  PetscPrintf(ctx->comm, "Elements: %d Nodes: %d Edges: %d\n", totalElements, totalNodes, totalEdges);
  CHKMEMQ;

  /* Replace old grid with refined grid */
  ierr = GridFinalizeBC(oldGrid);                                                                         CHKERRQ(ierr);
  ierr = GridDestroy(oldGrid);                                                                            CHKERRQ(ierr);
  ctx->grid = grid;
  PetscFunctionReturn(0);
}

/*------------------------------------------------- Grid Setup -------------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "HeatSetupGrid"
/*@
  HeatSetupGrid - This function sets all the functions,
  operators , and boundary conditions for the problem.
  It also sets the parameters associated with the fields.

  Collective on Grid

  Input Parameters:
+ grid - The grid
- ctx  - A HeatContext

  Options Database Keys:
. use_laplacian - Use the Laplacian in stead of the Rate-of-Strain tensor

  Level: intermediate

.seealso HeatCreateGrid()
@*/
int HeatSetupGrid(HeatContext ctx) {
  Grid grid  = ctx->grid;
  int  field = 0;
  int  ierr;

  PetscFunctionBegin;
  /* Setup Fields */
  ierr = GridSetFieldName(grid, field, "Temperature");                                                    CHKERRQ(ierr);

  /* Setup Problem */
  ierr = GridSetActiveField(grid, field);                                                                 CHKERRQ(ierr);

  /* Setup Rhs */
  ierr = HeatSetupRhsFunction(grid, ctx);                                                                 CHKERRQ(ierr);

  /* Setup Jacobian */
  ierr = GridSetMatOperator(grid, field, field, LAPLACIAN, -1.0, PETSC_FALSE, PETSC_NULL);                CHKERRQ(ierr);

  /* Setup Dirchlet boundary conditions */
  ierr = HeatSetupBC(grid, ctx);                                                                          CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatSetupRhsFunction"
/*@ HeatSetupRhsFunction
  HeatSetupRhsFunction - This function chooses a forcing  function for the problem.

  Collective on Grid

  Input Parameters:
+ grid - The grid
- ctx  - A HeatContext

  Level: intermediate

  Options Database Keys:

.seealso HeatSetupGrid
@*/
int HeatSetupRhsFunction(Grid grid, HeatContext ctx) {
  int field = 0;
  int ierr;

  PetscFunctionBegin;
  ierr = GridAddRhsFunction(grid, field, RhsFunction, 1.0);                                               CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatSetupBC"
/*@ HeatSetupBC
  HeatSetupBC - This function chooses boundary conditions for the problem.

  Collective on Grid

  Input Parameters:
+ grid - The grid
- ctx  - A HeatContext

  Level: intermediate

  Options Database Keys:
. bc_reduce - Explicitly reduce the system using boundary conditions

.seealso HeatSetupGrid()
@*/
int HeatSetupBC(Grid grid, HeatContext ctx) {
  int        field   = 0;
  PetscTruth reduceSystem;
  int        ierr;

  PetscFunctionBegin;
  ierr = PetscOptionsHasName(PETSC_NULL, "-bc_reduce", &reduceSystem);                                    CHKERRQ(ierr);
  switch(ctx->dim) {
  case 1:
    ierr = GridSetBC(grid, TOP_BD,    field, SolutionFunction, reduceSystem);                             CHKERRQ(ierr);
    ierr = GridAddBC(grid, BOTTOM_BD, field, SolutionFunction, reduceSystem);                             CHKERRQ(ierr);
    break;
  case 2:
    ierr = GridSetBC(grid, OUTER_BD, field, SolutionFunction, reduceSystem);                              CHKERRQ(ierr);
    break;
  default:
    SETERRQ1(PETSC_ERR_SUP, "Do not support meshes of dimension %d", ctx->dim);
  }
  ierr = GridSetBCContext(grid, ctx);                                                                     CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/*------------------------------------------------- SLES Setup -------------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "HeatCreateStructures"
int HeatCreateStructures(HeatContext ctx) {
  int size;
  int ierr;

  PetscFunctionBegin;
  /* Create solution and exact solution vectors */
  ierr = GVecCreate(ctx->grid, &ctx->u);                                                                  CHKERRQ(ierr);
  ierr = PetscObjectSetName((PetscObject) ctx->u, "Solution");                                            CHKERRQ(ierr);
  ierr = GVecDuplicate(ctx->u, &ctx->uExact);                                                             CHKERRQ(ierr);
  ierr = PetscObjectSetName((PetscObject) ctx->uExact, "ExactSolution");                                  CHKERRQ(ierr);
  /* Create the integrator */
  ierr = GTSCreate(ctx->grid, ctx, &ctx->ts);                                                             CHKERRQ(ierr);
  ierr = GTSSetTimeDependence(ctx->ts, 0, PETSC_TRUE);                                                    CHKERRQ(ierr);
  ierr = TSSetSolution(ctx->ts, ctx->u);                                                                  CHKERRQ(ierr);
  ierr = TSSetType(ctx->ts, "gbeuler");                                                                   CHKERRQ(ierr);
  ierr = TSSetFromOptions(ctx->ts);                                                                       CHKERRQ(ierr);

  ierr = VecGetSize(ctx->u, &size);                                                                       CHKERRQ(ierr);
  PetscPrintf(ctx->comm, "Variables: %d\n", size);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatDestroyTSMonitor"
int HeatDestroyTSMonitor(void *ctx) {
  GTSErrorMonitorCtx *monCtx = (GTSErrorMonitorCtx *) ctx;
  int                 ierr;

  PetscFunctionBegin;
  ierr = PetscViewerDestroy(monCtx->error_viewer);                                                        CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatSetupTS"
int HeatSetupTS(TS ts, HeatContext ctx) {
  GTSErrorMonitorCtx *monCtx = &ctx->monitorCtx;
  PetscViewer         v;
  PetscDraw           draw;
  char                viewerType[256];
  PetscTruth          opt;
  int                 ierr;

  PetscFunctionBegin;
  /* Setup convergence monitors */
  ierr = PetscOptionsGetString(PETSC_NULL, "-error_viewer", viewerType, 255, &opt);                       CHKERRQ(ierr);
  if (opt) {
    ierr = PetscStrcasecmp(viewerType, "draw", &opt);                                                     CHKERRQ(ierr);
    if (opt == PETSC_TRUE) {
      v    = PETSC_VIEWER_DRAW_(ctx->comm);
      ierr = PetscViewerSetFormat(v, PETSC_VIEWER_DRAW_LG);                                               CHKERRQ(ierr);
      ierr = PetscViewerDrawGetDraw(v, 0, &draw);                                                         CHKERRQ(ierr);
      ierr = PetscDrawSetTitle(draw, "Error");                                                            CHKERRQ(ierr);
    } else {
      ierr = PetscStrcasecmp(viewerType, "text", &opt);                                                   CHKERRQ(ierr);
      if (opt == PETSC_TRUE) {
        v    = PETSC_VIEWER_STDOUT_(ctx->comm);
        ierr = PetscObjectReference((PetscObject) v);                                                     CHKERRQ(ierr);
      } else {
        SETERRQ1(PETSC_ERR_ARG_WRONG, "Invalid error viewer: %s", viewerType);
      }
    }
  } else {
    ierr = PetscViewerASCIIOpen(ctx->comm, "/dev/null", &v);                                              CHKERRQ(ierr);
  }
  monCtx->error_viewer      = v;
  monCtx->solutionFunc      = HeatSolutionFunction;
  monCtx->ctx               = ctx;
  monCtx->solution          = ctx->uExact;
  monCtx->norm_error_viewer = PETSC_VIEWER_STDOUT_(ctx->comm);
  /*ierr = TSSetMonitor(ts, GTSErrorMonitor, monCtx, HeatDestroyTSMonitor);                                 CHKERRQ(ierr);*/
  ierr = TSSetMonitor(ts, GTSErrorMonitor, monCtx, PETSC_NULL);                                           CHKERRQ(ierr);

  ierr = TSSetUpdate(ts, HeatUpdateGTS);                                                                  CHKERRQ(ierr);
  ierr = TSSetUp(ts);                                                                                     CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatDestroyKSPMonitor"
int HeatDestroyKSPMonitor(void *ctx) {
  GVecErrorKSPMonitorCtx *monCtx = (GVecErrorKSPMonitorCtx *) ctx;
  int                     ierr;

  PetscFunctionBegin;
  ierr = PetscViewerDestroy(monCtx->error_viewer);                                                        CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatSetupKSP"
int HeatSetupKSP(KSP ksp, HeatContext ctx) {
  GVecErrorKSPMonitorCtx *monCtx = &ctx->kspMonitorCtx;
  PetscViewer             v;
  PetscDraw               draw;
  PetscTruth              opt;
  int                     ierr;

  PetscFunctionBegin;
  /* Setup convergence monitors */
  ierr = PetscOptionsHasName(PETSC_NULL, "-error_viewer", &opt);                                          CHKERRQ(ierr);
  if (opt == PETSC_TRUE) {
    v    = PETSC_VIEWER_DRAW_(ctx->comm);
    ierr = PetscViewerSetFormat(v, PETSC_VIEWER_DRAW_LG);                                                 CHKERRQ(ierr);
    ierr = PetscViewerDrawGetDraw(v, 0, &draw);                                                           CHKERRQ(ierr);
    ierr = PetscDrawSetTitle(draw, "Error");                                                              CHKERRQ(ierr);
  } else {
    v    = PETSC_VIEWER_STDOUT_(ctx->comm);
  }
  monCtx->error_viewer      = v;
  monCtx->solution          = ctx->uExact;
  monCtx->norm_error_viewer = PETSC_VIEWER_STDOUT_(ctx->comm);
  /*ierr = KSPSetMonitor(ksp, GVecErrorKSPMonitor, monCtx, HeatDestroyKSPMonitor);                          CHKERRQ(ierr);*/
  ierr = KSPSetMonitor(ksp, GVecErrorKSPMonitor, monCtx, PETSC_NULL);                                     CHKERRQ(ierr);
  PetscFunctionReturn(0);
}


#undef  __FUNCT__
#define __FUNCT__ "HeatSetupPC"
int HeatSetupPC(PC pc, HeatContext ctx) {
  PetscFunctionBegin;
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatSetupStructures"
int HeatSetupStructures(HeatContext ctx) {
  SLES       sles;
  KSP        ksp;
  PC         pc;
  GTSContext oldCtx;
  double     time;
  int        field  = 0;
  int        ierr;

  PetscFunctionBegin;
  /* Setup the integrator */
  ierr = HeatSetupTS(ctx->ts, ctx);                                                                       CHKERRQ(ierr);
  /* Setup the linear solver */
  ierr = TSGetSLES(ctx->ts, &sles);                                                                       CHKERRQ(ierr);
  ierr = SLESGetKSP(sles, &ksp);                                                                          CHKERRQ(ierr);
  ierr = HeatSetupKSP(ksp, ctx);                                                                          CHKERRQ(ierr);
  ierr = SLESGetPC(sles, &pc);                                                                            CHKERRQ(ierr);
  ierr = HeatSetupPC(pc, ctx);                                                                            CHKERRQ(ierr);
  /* Evaluate the exact solution */
  ierr = TSGetTime(ctx->ts, &time);                                                                       CHKERRQ(ierr);
  ierr = GTSCreateContext(ctx->ts, time, (PetscObject) ctx, (PetscObject *) &oldCtx);                     CHKERRQ(ierr);
  ierr = GVecEvaluateFunction(ctx->uExact, 1, &field, SolutionFunction, 1.0, ctx);                        CHKERRQ(ierr);
  ierr = GTSDestroyContext(ctx->ts, (PetscObject) ctx, (PetscObject) oldCtx);                             CHKERRQ(ierr);
  /* Set the inital solution to be the exact solution */
  ierr = VecCopy(ctx->uExact, ctx->u);                                                                    CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatDestroyStructures"
int HeatDestroyStructures(HeatContext ctx) {
  int ierr;

  PetscFunctionBegin;
  ierr = GTSDestroy(ctx->ts);                                                                             CHKERRQ(ierr);
  ierr = GVecDestroy(ctx->u);                                                                             CHKERRQ(ierr);
  ierr = GVecDestroy(ctx->uExact);                                                                        CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/*----------------------------------------------- Sanity Checks ------------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "HeatCheckSolution"
int HeatCheckSolution(HeatContext ctx, double time, GVec u, const char type[]) {
  GMat         A, M;
  void        *matCtx;
  MatStructure str;
  GVec         u_t, f, r;
  GTSContext   oldCtx;
  PetscScalar  norm;
  PetscScalar  one = 1.0, minusOne = -1.0;
  int          T   = 0;
  int          ierr;

  PetscFunctionBegin;
  ierr = GVecDuplicate(u, &u_t);                                                                          CHKERRQ(ierr);
  ierr = GVecDuplicate(u, &f);                                                                            CHKERRQ(ierr);
  ierr = GVecDuplicate(u, &r);                                                                            CHKERRQ(ierr);
  /* Evaluate f */
  ierr = GTSEvaluateRhs(ctx->ts, time, u, f, (PetscObject) ctx);                                          CHKERRQ(ierr);
  ierr = GTSRhsBC(ctx->ts, f, ctx);                                                                       CHKERRQ(ierr);
  /* Evaluate u_t */
  ierr = GTSCreateContext(ctx->ts, time, (PetscObject) ctx, (PetscObject *) &oldCtx);                     CHKERRQ(ierr);
  ierr = GVecEvaluateFunctionGalerkin(u_t, 1, &T, SolutionTimeDerivative, 1.0, ctx);                      CHKERRQ(ierr);
  ierr = GTSDestroyContext(ctx->ts, (PetscObject) ctx, (PetscObject) oldCtx);                             CHKERRQ(ierr);
  /* Evaluate A */
  ierr = TSGetRHSMatrix(ctx->ts, &A, &M, &matCtx);                                                        CHKERRQ(ierr);
  ierr = GTSEvaluateSystemMatrix(ctx->ts, time, &A, &M, &str, (PetscObject) matCtx);                      CHKERRQ(ierr);
  /* Evaluate r = f - u_t - A u */
  ierr = MatMult(A, u, r);                                                                                CHKERRQ(ierr);
  ierr = VecAXPY(&one, u_t, r);                                                                           CHKERRQ(ierr);
  ierr = GTSRhsBC(ctx->ts, r, ctx);                                                                       CHKERRQ(ierr);
  ierr = VecAYPX(&minusOne, f, r);                                                                        CHKERRQ(ierr);
  ierr = VecNorm(r, NORM_2, &norm);                                                                       CHKERRQ(ierr);
  PetscPrintf(ctx->comm, "Residual of the %s solution: %g\n", type, norm);
  ierr = GVecDestroy(u_t);                                                                                CHKERRQ(ierr);
  ierr = GVecDestroy(f);                                                                                  CHKERRQ(ierr);
  ierr = GVecDestroy(r);                                                                                  CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/*----------------------------------------------- Problem Callbacks --------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "HeatSolutionFunction"
/*@
  HeatSolutionFunction - This function is the solution function for the problem.

  Not collective

  Input Parameters:
+ n      - The number of points
. comp   - The number of components
. x,y,z  - The points
. values - The output
- ctx    - A NavierStokesContext

  Level: beginner

  Note:
  The solution is exactly SolutionFunction

.keywords solution
.seealso RhsFunction(), SolutionFunction()
@*/
int HeatSolutionFunction(Vec u, void *ctx) {
  int field = 0;
  int ierr;

  PetscFunctionBegin;
  ierr = GVecEvaluateFunction(u, 1, &field, SolutionFunction, 1.0, ctx);                                    CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "SolutionFunction"
/*@
  SolutionFunction - This function is the solution function for the problem.

  Not collective

  Input Parameters:
+ n      - The number of points
. comp   - The number of components
. x,y,z  - The points
. values - The output
- ctx    - A HeatContext

  Level: beginner

  Note:
$  The solution for one dimension  is u = (x^2 \hat x) e^{-k^2 t}.
$  The solution for two dimensions is u = (x^2 \hat x - 2 x y \hat y) e^{-k^2 t}.

.keywords solution
.seealso RhsFunction()
@*/
int SolutionFunction(int n, int comp, double *x, double *y, double *z, PetscScalar *values, void *ctx) {
  GTSContext GTSctx;
  double     k, t, w;
  int        i;
  int        ierr;

  PetscFunctionBegin;
  ierr = PetscObjectQuery((PetscObject) ctx, "GTSContext", (PetscObject *) &GTSctx);                      CHKERRQ(ierr);
  t    = GTSctx->t;
  k    = ((HeatContext) ctx)->decayExponent;
  w    = exp(-k*k*t);
  switch(comp) {
  case 1:
    for(i = 0; i < n; i++) {
      values[i] = x[i]*x[i]*w;
    }
    break;
  case 2:
    for(i = 0; i < n; i++) {
      values[i*2+0] = x[i]*x[i]*w;
      values[i*2+1] = -2.0*x[i]*y[i]*w;
    }
    break;
  default:
    SETERRQ1(PETSC_ERR_ARG_WRONG, "Invalid number of components %d", comp);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "SolutionTimeDerivative"
/*@
  SolutionTimeDerivative - This function is the time derivative of the solution function for the problem.

  Not collective

  Input Parameters:
+ n      - The number of points
. comp   - The number of components
. x,y,z  - The points
. values - The output
- ctx    - A HeatContext

  Level: beginner

  Note:
$  The solution for one dimension  is u = -(k^2 x^2 \hat x) e^{-k^2 t}.
$  The solution for two dimensions is u = -(k^2 x^2 \hat x - 2 k^2 x y \hat y) e^{-k^2 t}.

.keywords solution, derivative
.seealso RhsFunction()
@*/
int SolutionTimeDerivative(int n, int comp, double *x, double *y, double *z, PetscScalar *values, void *ctx) {
  GTSContext GTSctx;
  double     k, t, w;
  int        i;
  int        ierr;

  PetscFunctionBegin;
  ierr = PetscObjectQuery((PetscObject) ctx, "GTSContext", (PetscObject *) &GTSctx);                      CHKERRQ(ierr);
  t    = GTSctx->t;
  k    = ((HeatContext) ctx)->decayExponent;
  w    = exp(-k*k*t);
  switch(comp) {
  case 1:
    for(i = 0; i < n; i++) {
      values[i] = -k*k*x[i]*x[i]*w;
    }
    break;
  case 2:
    for(i = 0; i < n; i++) {
      values[i*2+0] = -k*k*x[i]*x[i]*w;
      values[i*2+1] = 2.0*k*k*x[i]*y[i]*w;
    }
    break;
  default:
    SETERRQ1(PETSC_ERR_ARG_WRONG, "Invalid number of components %d", comp);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "RhsFunction"
/*@
  RhsFunction - This function is the forcing function for the problem.

  Not collective

  Input Parameters:
+ n      - The number of points
. comp   - The number of components
. x,y,z  - The points
. values - The output
- ctx    - A HeatContext

  Level: beginner

  Note:
$  The rhs for one dimension  is u_t - \Delta u = -(k^2 x^2 + 2 \hat x) e^{-k^2 t}.
$  The rhs for two dimensions is u_t - \Delta u = -(k^2 x^2 + 2 \hat x - 2 k^2 x y \hat y) e^{-k^2 t}.

.keywords rhs
.seealso SolutionFunction()
@*/
int RhsFunction(int n, int comp, double *x, double *y, double *z, PetscScalar *values, void *ctx) {
  GTSContext GTSctx;
  double     k, t, w;
  int        i;
  int        ierr;

  PetscFunctionBegin;
  ierr = PetscObjectQuery((PetscObject) ctx, "GTSContext", (PetscObject *) &GTSctx);                      CHKERRQ(ierr);
  t    = GTSctx->t;
  k    = ((HeatContext) ctx)->decayExponent;
  w    = exp(-k*k*t);
  switch(comp) {
  case 1:
    for(i = 0; i < n; i++) {
      values[i] = -(k*k*x[i]*x[i] + 2.0)*w;
    }
    break;
  case 2:
    for(i = 0; i < n; i++) {
      values[i*2+0] = -(k*k*x[i]*x[i] + 2.0)*w;
      values[i*2+1] = 2.0*k*k*x[i]*y[i]*w;
    }
    break;
  default:
    SETERRQ1(PETSC_ERR_ARG_WRONG, "Invalid number of components %d", comp);
  }
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatUpdateGTS"
int HeatUpdateGTS(TS ts, double time, double *dt) {
  HeatContext ctx;
  GTSContext  oldCtx;
  int         field = 0;
  int         ierr;

  PetscFunctionBegin;
  ierr = TSGetRHSMatrix(ts, PETSC_NULL, PETSC_NULL, (void **) &ctx);                                      CHKERRQ(ierr);
  ierr = GTSCreateContext(ts, time, (PetscObject) ctx, (PetscObject *) &oldCtx);                          CHKERRQ(ierr);
  ierr = GVecEvaluateFunction(ctx->uExact, 1, &field, SolutionFunction, 1.0, ctx);                        CHKERRQ(ierr);
  ierr = GTSDestroyContext(ts, (PetscObject) ctx, (PetscObject) oldCtx);                                  CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/*----------------------------------------------- Main Computation ---------------------------------------------------*/
#undef  __FUNCT__
#define __FUNCT__ "HeatSolve"
int HeatSolve(HeatContext ctx, double *time) {
  int steps;
  int ierr;

  PetscFunctionBegin;
  ierr = TSStep(ctx->ts, &steps, time);                                                                   CHKERRQ(ierr);

  /* Show solution */
  ierr = GVecViewFromOptions(ctx->uExact, "Exact Solution");                                              CHKERRQ(ierr);
  ierr = GVecViewFromOptions(ctx->u,      "Solution");                                                    CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef  __FUNCT__
#define __FUNCT__ "HeatComputeField"
int HeatComputeField(HeatContext ctx) {
  double time = 0.0;
  double dt   = 0.0;
  int    loop;
  int    ierr;

  /* Get command-line options */
  ierr = HeatContextSetup(ctx);                                                                           CHKERRQ(ierr);

  for(loop = 0; loop < ctx->numLoops; loop++) {
    if (loop >= ctx->refineStep) {
      ierr = HeatRefineGrid(ctx);                                                                         CHKERRQ(ierr);
    }

    /* Setup problem */
    ierr = HeatSetupGrid(ctx);                                                                            CHKERRQ(ierr);
    ierr = HeatCreateStructures(ctx);                                                                     CHKERRQ(ierr);
    ierr = HeatSetupStructures(ctx);                                                                      CHKERRQ(ierr);

    /* Check the exact solution */
    ierr = HeatCheckSolution(ctx, time, ctx->uExact, "exact");                                            CHKERRQ(ierr);

    /* Solve system */
    ierr = HeatSolve(ctx, &time);                                                                         CHKERRQ(ierr);

    /* Check the exact solution */
    ierr = HeatUpdateGTS(ctx->ts, time, &dt);                                                             CHKERRQ(ierr);
    ierr = HeatCheckSolution(ctx, time, ctx->uExact, "exact");                                            CHKERRQ(ierr);

    /* Check the computed solution */
    ierr = HeatCheckSolution(ctx, time, ctx->u, "computed");                                              CHKERRQ(ierr);

    /* Cleanup */
    ierr = HeatDestroyStructures(ctx);                                                                    CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}
